@@ -1,5 +1,6 @@
 {-# LANGUAGE ExistentialQuantification, FlexibleInstances, GeneralizedNewtypeDeriving,
-             MultiParamTypeClasses, TypeSynonymInstances, DeriveDataTypeable #-}
+             MultiParamTypeClasses, TypeSynonymInstances, DeriveDataTypeable,
+             LambdaCase, NamedFieldPuns, DeriveTraversable #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -22,26 +23,29 @@ module XMonad.Core (
     XConf(..), XConfig(..), LayoutClass(..),
     Layout(..), readsLayout, Typeable, Message,
     SomeMessage(..), fromMessage, LayoutMessages(..),
-    StateExtension(..), ExtensionClass(..),
+    StateExtension(..), ExtensionClass(..), ConfExtension(..),
     runX, catchX, userCode, userCodeDef, io, catchIO, installSignalHandlers, uninstallSignalHandlers,
     withDisplay, withWindowSet, isRoot, runOnWorkspaces,
-    getAtom, spawn, spawnPID, xfork, recompile, trace, whenJust, whenX,
-    getXMonadDir, getXMonadCacheDir, getXMonadDataDir, stateFileName,
+    getAtom, spawn, spawnPID, xfork, xmessage, recompile, trace, whenJust, whenX,
+    getXMonadDir, getXMonadCacheDir, getXMonadDataDir, stateFileName, binFileName,
     atom_WM_STATE, atom_WM_PROTOCOLS, atom_WM_DELETE_WINDOW, atom_WM_TAKE_FOCUS, withWindowAttributes,
-    ManageHook, Query(..), runQuery
+    ManageHook, Query(..), runQuery, Directories'(..), Directories, getDirectories,
   ) where
 
 import XMonad.StackSet hiding (modify)
 
 import Prelude
-import Control.Exception.Extensible (fromException, try, bracket, throw, finally, SomeException(..))
-import qualified Control.Exception.Extensible as E
-import Control.Applicative(Applicative, pure, (<$>), (<*>))
+import Control.Exception (fromException, try, bracket, bracket_, throw, finally, SomeException(..))
+import qualified Control.Exception as E
+import Control.Applicative ((<|>), empty)
 import Control.Monad.Fail
 import Control.Monad.State
 import Control.Monad.Reader
 import Data.Semigroup
-import Data.Default
+import Data.Traversable (for)
+import Data.Time.Clock (UTCTime)
+import Data.Default.Class
+import Data.List (isInfixOf)
 import System.FilePath
 import System.IO
 import System.Info
@@ -58,8 +62,6 @@ import Graphics.X11.Xlib.Extras (getWindowAttributes, WindowAttributes, Event)
 import Data.Typeable
 import Data.List ((\\))
 import Data.Maybe (isJust,fromMaybe)
-import Data.Monoid hiding ((<>))
-import System.Environment (lookupEnv)
 
 import qualified Data.Map as M
 import qualified Data.Set as S
@@ -93,8 +95,8 @@ data XConf = XConf
     , mousePosition :: !(Maybe (Position, Position))
                                       -- ^ position of the mouse according to
                                       -- the event currently being processed
-    , currentEvent :: !(Maybe Event)
-                                      -- ^ event currently being processed
+    , currentEvent :: !(Maybe Event)  -- ^ event currently being processed
+    , directories  :: !Directories    -- ^ directories to use
     }
 
 -- todo, better name
@@ -122,6 +124,11 @@ data XConfig l = XConfig
     , rootMask           :: !EventMask           -- ^ The root events that xmonad is interested in
     , handleExtraArgs    :: !([String] -> XConfig Layout -> IO (XConfig Layout))
                                                  -- ^ Modify the configuration, complain about extra arguments etc. with arguments that are not handled by default
+    , extensibleConf     :: !(M.Map TypeRep ConfExtension)
+                                                 -- ^ Stores custom config information.
+                                                 --
+                                                 -- The module "XMonad.Util.ExtensibleConf" in xmonad-contrib
+                                                 -- provides additional information and a simple interface for using this.
     }
 
 
@@ -135,7 +142,8 @@ type WorkspaceId = String
 newtype ScreenId    = S Int deriving (Eq,Ord,Show,Read,Enum,Num,Integral,Real)
 
 -- | The 'Rectangle' with screen dimensions
-data ScreenDetail   = SD { screenRect :: !Rectangle } deriving (Eq,Show, Read)
+newtype ScreenDetail = SD { screenRect :: Rectangle }
+    deriving (Eq,Show, Read)
 
 ------------------------------------------------------------------------
 
@@ -148,7 +156,7 @@ data ScreenDetail   = SD { screenRect :: !Rectangle } deriving (Eq,Show, Read)
 -- instantiated on 'XConf' and 'XState' automatically.
 --
 newtype X a = X (ReaderT XConf (StateT XState IO) a)
-    deriving (Functor, Monad, MonadFail, MonadIO, MonadState XState, MonadReader XConf, Typeable)
+    deriving (Functor, Monad, MonadFail, MonadIO, MonadState XState, MonadReader XConf)
 
 instance Applicative X where
   pure = return
@@ -255,14 +263,18 @@ readsLayout (Layout l) s = [(Layout (asTypeOf x l), rs) | (x, rs) <- reads s]
 -- | Every layout must be an instance of 'LayoutClass', which defines
 -- the basic layout operations along with a sensible default for each.
 --
--- Minimal complete definition:
+-- All of the methods have default implementations, so there is no
+-- minimal complete definition.  They do, however, have a dependency
+-- structure by default; this is something to be aware of should you
+-- choose to implement one of these methods.  Here is how a minimal
+-- complete definition would look like if we did not provide any default
+-- implementations:
 --
--- * 'runLayout' || (('doLayout' || 'pureLayout') && 'emptyLayout'), and
+-- * 'runLayout' || (('doLayout' || 'pureLayout') && 'emptyLayout')
 --
 -- * 'handleMessage' || 'pureMessage'
 --
--- You should also strongly consider implementing 'description',
--- although it is not required.
+-- * 'description'
 --
 -- Note that any code which /uses/ 'LayoutClass' methods should only
 -- ever call 'runLayout', 'handleMessage', and 'description'!  In
@@ -271,7 +283,7 @@ readsLayout (Layout l) s = [(Layout (asTypeOf x l), rs) | (x, rs) <- reads s]
 -- 'runLayout', 'handleMessage', and so on.  This ensures that the
 -- proper methods will be used, regardless of the particular methods
 -- that any 'LayoutClass' instance chooses to define.
-class Show (layout a) => LayoutClass layout a where
+class (Show (layout a), Typeable layout) => LayoutClass layout a where
 
     -- | By default, 'runLayout' calls 'doLayout' if there are any
     --   windows to be laid out, and 'emptyLayout' otherwise.  Most
@@ -373,12 +385,12 @@ instance Message Event
 -- layouts) should consider handling.
 data LayoutMessages = Hide              -- ^ sent when a layout becomes non-visible
                     | ReleaseResources  -- ^ sent when xmonad is exiting or restarting
-    deriving (Typeable, Eq)
+    deriving Eq
 
 instance Message LayoutMessages
 
 -- ---------------------------------------------------------------------
--- Extensible state
+-- Extensible state/config
 --
 
 -- | Every module must make the data it wants to store
@@ -386,6 +398,7 @@ instance Message LayoutMessages
 --
 -- Minimal complete definition: initialValue
 class Typeable a => ExtensionClass a where
+    {-# MINIMAL initialValue #-}
     -- | Defines an initial value for the state extension
     initialValue :: a
     -- | Specifies whether the state extension should be
@@ -404,6 +417,9 @@ data StateExtension =
   | forall a. (Read a, Show a, ExtensionClass a) => PersistentExtension a
     -- ^ Persistent extension
 
+-- | Existential type to store a config extension.
+data ConfExtension = forall a. Typeable a => ConfExtension a
+
 -- ---------------------------------------------------------------------
 -- | General utilities
 --
@@ -439,6 +455,16 @@ xfork x = io . forkProcess . finally nullStdin $ do
         dupTo fd stdInput
         closeFd fd
 
+-- | Use @xmessage@ to show information to the user.
+xmessage :: MonadIO m => String -> m ()
+xmessage msg = void . xfork $ do
+    executeFile "xmessage" True
+        [ "-default", "okay"
+        , "-xrm", "*international:true"
+        , "-xrm", "*fontSet:-*-fixed-medium-r-normal-*-18-*-*-*-*-*-*-*,-*-fixed-*-*-*-*-18-*-*-*-*-*-*-*,-*-*-*-*-*-*-18-*-*-*-*-*-*-*"
+        , msg
+        ] Nothing
+
 -- | This is basically a map function, running a function in the 'X' monad on
 -- each workspace with the output of that function being the modified workspace.
 runOnWorkspaces :: (WindowSpace -> X WindowSpace) -> X ()
@@ -449,137 +475,254 @@ runOnWorkspaces job = do
              $ current ws : visible ws
     modify $ \s -> s { windowset = ws { current = c, visible = v, hidden = h } }
 
--- | Return the path to the xmonad configuration directory.  This
--- directory is where user configuration files are stored (e.g, the
--- xmonad.hs file).  You may also create a @lib@ subdirectory in the
--- configuration directory and the default recompile command will add
--- it to the GHC include path.
+-- | All the directories that xmonad will use.  They will be used for
+-- the following purposes:
 --
--- Several directories are considered.  In order of
--- preference:
+-- * @dataDir@: This directory is used by XMonad to store data files
+-- such as the run-time state file.
 --
---   1. The directory specified in the @XMONAD_CONFIG_DIR@ environment variable.
---   2. The @~\/.xmonad@ directory.
---   3. The @XDG_CONFIG_HOME/xmonad@ directory.
+-- * @cfgDir@: This directory is where user configuration files are
+-- stored (e.g, the xmonad.hs file).  You may also create a @lib@
+-- subdirectory in the configuration directory and the default recompile
+-- command will add it to the GHC include path.
 --
--- The first directory that exists will be used.  If none of the
--- directories exist then (1) will be used if it is set, otherwise (2)
--- will be used.  Either way, a directory will be created if necessary.
-getXMonadDir :: MonadIO m => m String
-getXMonadDir =
-    findFirstDirWithEnv "XMONAD_CONFIG_DIR"
-      [ getAppUserDataDirectory "xmonad"
-      , getXDGDirectory XDGConfig "xmonad"
-      ]
+-- * @cacheDir@: This directory is used to store temporary files that
+-- can easily be recreated such as the configuration binary and any
+-- intermediate object files generated by GHC.
+-- Also, the XPrompt history file goes here.
+--
+-- For how these directories are chosen, see 'getDirectories'.
+--
+data Directories' a = Directories
+    { dataDir  :: !a
+    , cfgDir   :: !a
+    , cacheDir :: !a
+    }
+    deriving (Show, Functor, Foldable, Traversable)
 
--- | Return the path to the xmonad cache directory.  This directory is
--- used to store temporary files that can easily be recreated.  For
--- example, the XPrompt history file.
---
--- Several directories are considered.  In order of preference:
---
---   1. The directory specified in the @XMONAD_CACHE_DIR@ environment variable.
---   2. The @~\/.xmonad@ directory.
---   3. The @XDG_CACHE_HOME/xmonad@ directory.
---
--- The first directory that exists will be used.  If none of the
--- directories exist then (1) will be used if it is set, otherwise (2)
--- will be used.  Either way, a directory will be created if necessary.
-getXMonadCacheDir :: MonadIO m => m String
-getXMonadCacheDir =
-    findFirstDirWithEnv "XMONAD_CACHE_DIR"
-      [ getAppUserDataDirectory "xmonad"
-      , getXDGDirectory XDGCache "xmonad"
-      ]
+-- | Convenient type alias for the most common case in which one might
+-- want to use the 'Directories' type.
+type Directories = Directories' FilePath
 
--- | Return the path to the xmonad data directory.  This directory is
--- used by XMonad to store data files such as the run-time state file
--- and the configuration binary generated by GHC.
+-- | Build up the 'Dirs' that xmonad will use.  They are chosen as
+-- follows:
 --
--- Several directories are considered.  In order of preference:
+-- 1. If all three of xmonad's environment variables (@XMONAD_DATA_DIR@,
+--    @XMONAD_CONFIG_DIR@, and @XMONAD_CACHE_DIR@) are set, use them.
+-- 2. If there is a build script called @build@ or configuration
+--    @xmonad.hs@ in @~\/.xmonad@, set all three directories to
+--    @~\/.xmonad@.
+-- 3. Otherwise, use the @xmonad@ directory in @XDG_DATA_HOME@,
+--    @XDG_CONFIG_HOME@, and @XDG_CACHE_HOME@ (or their respective
+--    fallbacks).  These directories are created if necessary.
 --
---   1. The directory specified in the @XMONAD_DATA_DIR@ environment variable.
---   2. The @~\/.xmonad@ directory.
---   3. The @XDG_DATA_HOME/xmonad@ directory.
+-- The xmonad configuration file (or the build script, if present) is
+-- always assumed to be in @cfgDir@.
 --
--- The first directory that exists will be used.  If none of the
--- directories exist then (1) will be used if it is set, otherwise (2)
--- will be used.  Either way, a directory will be created if necessary.
-getXMonadDataDir :: MonadIO m => m String
-getXMonadDataDir =
-    findFirstDirWithEnv "XMONAD_DATA_DIR"
-      [ getAppUserDataDirectory "xmonad"
-      , getXDGDirectory XDGData "xmonad"
-      ]
-
--- | Helper function that will find the first existing directory and
--- return its path.  If none of the directories can be found, create
--- and return the first from the list.  If the list is empty this
--- function returns the historical @~\/.xmonad@ directory.
-findFirstDirOf :: MonadIO m => [IO FilePath] -> m FilePath
-findFirstDirOf []        = findFirstDirOf [getAppUserDataDirectory "xmonad"]
-findFirstDirOf possibles = do
-    found <- go possibles
-
-    case found of
-      Just path -> return path
-      Nothing   -> do
-        primary <- io (head possibles)
-        io (createDirectoryIfMissing True primary)
-        return primary
-
+getDirectories :: IO Directories
+getDirectories = xmEnvDirs <|> xmDirs <|> xdgDirs
   where
-    go []     = return Nothing
-    go (x:xs) = do
-      dir    <- io x
-      exists <- io (doesDirectoryExist dir)
-      if exists then return (Just dir) else go xs
+    -- | Check for xmonad's environment variables first
+    xmEnvDirs :: IO Directories
+    xmEnvDirs = do
+        let xmEnvs = Directories{ dataDir  = "XMONAD_DATA_DIR"
+                                , cfgDir   = "XMONAD_CONFIG_DIR"
+                                , cacheDir = "XMONAD_CACHE_DIR"
+                                }
+        maybe empty pure . sequenceA =<< traverse getEnv xmEnvs
 
--- | Simple wrapper around @findFirstDirOf@ that allows the primary
--- path to be specified by an environment variable.
-findFirstDirWithEnv :: MonadIO m => String -> [IO FilePath] -> m FilePath
-findFirstDirWithEnv envName paths = do
-    envPath' <- io (getEnv envName)
+    -- | Check whether the config file or a build script is in the
+    -- @~\/.xmonad@ directory
+    xmDirs :: IO Directories
+    xmDirs = do
+        xmDir <- getAppUserDataDirectory "xmonad"
+        conf  <- doesFileExist $ xmDir </> "xmonad.hs"
+        build <- doesFileExist $ xmDir </> "build"
 
-    case envPath' of
-      Nothing      -> findFirstDirOf paths
-      Just envPath -> findFirstDirOf (return envPath:paths)
+        -- Place *everything* in ~/.xmonad if yes
+        guard $ conf || build
+        pure Directories{ dataDir = xmDir, cfgDir = xmDir, cacheDir = xmDir }
 
--- | Helper function to retrieve the various XDG directories.
--- This has been based on the implementation shipped with GHC version 8.0.1 or
--- higher. Put here to preserve compatibility with older GHC versions.
-getXDGDirectory :: XDGDirectory -> FilePath -> IO FilePath
-getXDGDirectory xdgDir suffix =
-  normalise . (</> suffix) <$>
-  case xdgDir of
-    XDGData   -> get "XDG_DATA_HOME"   ".local/share"
-    XDGConfig -> get "XDG_CONFIG_HOME" ".config"
-    XDGCache  -> get "XDG_CACHE_HOME"  ".cache"
+    -- | Use XDG directories as a fallback
+    xdgDirs :: IO Directories
+    xdgDirs =
+        for Directories{ dataDir = XdgData, cfgDir = XdgConfig, cacheDir = XdgCache }
+            $ \dir -> do d <- getXdgDirectory dir "xmonad"
+                         d <$ createDirectoryIfMissing True d
+
+-- | Return the path to the xmonad configuration directory.
+getXMonadDir :: X String
+getXMonadDir = asks (cfgDir . directories)
+{-# DEPRECATED getXMonadDir "Use `asks (cfgDir . directories)' instead." #-}
+
+-- | Return the path to the xmonad cache directory.
+getXMonadCacheDir :: X String
+getXMonadCacheDir = asks (cacheDir . directories)
+{-# DEPRECATED getXMonadCacheDir "Use `asks (cacheDir . directories)' instead." #-}
+
+-- | Return the path to the xmonad data directory.
+getXMonadDataDir :: X String
+getXMonadDataDir = asks (dataDir . directories)
+{-# DEPRECATED getXMonadDataDir "Use `asks (dataDir . directories)' instead." #-}
+
+binFileName, buildDirName :: Directories -> FilePath
+binFileName  Directories{ cacheDir } = cacheDir </> "xmonad-" <> arch <> "-" <> os
+buildDirName Directories{ cacheDir } = cacheDir </> "build-" <> arch <> "-" <> os
+
+errFileName, stateFileName :: Directories -> FilePath
+errFileName   Directories{ dataDir } = dataDir </> "xmonad.errors"
+stateFileName Directories{ dataDir } = dataDir </> "xmonad.state"
+
+srcFileName, libFileName :: Directories -> FilePath
+srcFileName Directories{ cfgDir } = cfgDir </> "xmonad.hs"
+libFileName Directories{ cfgDir } = cfgDir </> "lib"
+
+buildScriptFileName, stackYamlFileName :: Directories -> FilePath
+buildScriptFileName Directories{ cfgDir } = cfgDir </> "build"
+stackYamlFileName   Directories{ cfgDir } = cfgDir </> "stack.yaml"
+
+-- | Compilation method for xmonad configuration.
+data Compile = CompileGhc | CompileStackGhc FilePath | CompileScript FilePath
+    deriving (Show)
+
+-- | Detect compilation method by looking for known file names in xmonad
+-- configuration directory.
+detectCompile :: Directories -> IO Compile
+detectCompile dirs = tryScript <|> tryStack <|> useGhc
   where
-    get name fallback = do
-      env <- lookupEnv name
-      case env of
-        Nothing -> fallback'
-        Just path
-          | isRelative path -> fallback'
-          | otherwise -> return path
-      where
-        fallback' = (</> fallback) <$> getHomeDirectory
-data XDGDirectory = XDGData | XDGConfig | XDGCache
+    buildScript = buildScriptFileName dirs
+    stackYaml = stackYamlFileName dirs
 
--- | Get the name of the file used to store the xmonad window state.
-stateFileName :: (Functor m, MonadIO m) => m FilePath
-stateFileName = (</> "xmonad.state") <$> getXMonadDataDir
+    tryScript = do
+        guard =<< doesFileExist buildScript
+        isExe <- isExecutable buildScript
+        if isExe
+          then do
+            trace $ "XMonad will use build script at " <> show buildScript <> " to recompile."
+            pure $ CompileScript buildScript
+          else do
+            trace $ "XMonad will not use build script, because " <> show buildScript <> " is not executable."
+            trace $ "Suggested resolution to use it: chmod u+x " <> show buildScript
+            empty
 
--- | 'recompile force', recompile the xmonad configuration file when
--- any of the following apply:
+    tryStack = do
+        guard =<< doesFileExist stackYaml
+        canonStackYaml <- canonicalizePath stackYaml
+        trace $ "XMonad will use stack ghc --stack-yaml " <> show canonStackYaml <> " to recompile."
+        pure $ CompileStackGhc canonStackYaml
+
+    useGhc = do
+        trace $ "XMonad will use ghc to recompile, because neither "
+                <> show buildScript <> " nor " <> show stackYaml <> " exists."
+        pure CompileGhc
+
+    isExecutable f = E.catch (executable <$> getPermissions f) (\(SomeException _) -> return False)
+
+-- | Should we recompile xmonad configuration? Is it newer than the compiled
+-- binary?
+shouldCompile :: Directories -> Compile -> IO Bool
+shouldCompile dirs CompileGhc = do
+    libTs <- mapM getModTime . Prelude.filter isSource =<< allFiles (libFileName dirs)
+    srcT <- getModTime (srcFileName dirs)
+    binT <- getModTime (binFileName dirs)
+    if any (binT <) (srcT : libTs)
+        then True <$ trace "XMonad recompiling because some files have changed."
+        else False <$ trace "XMonad skipping recompile because it is not forced (e.g. via --recompile), and neither xmonad.hs nor any *.hs / *.lhs / *.hsc files in lib/ have been changed."
+  where
+    isSource = flip elem [".hs",".lhs",".hsc"] . takeExtension
+    allFiles t = do
+        let prep = map (t</>) . Prelude.filter (`notElem` [".",".."])
+        cs <- prep <$> E.catch (getDirectoryContents t) (\(SomeException _) -> return [])
+        ds <- filterM doesDirectoryExist cs
+        concat . ((cs \\ ds):) <$> mapM allFiles ds
+shouldCompile dirs CompileStackGhc{} = do
+    stackYamlT <- getModTime (stackYamlFileName dirs)
+    binT <- getModTime (binFileName dirs)
+    if binT < stackYamlT
+        then True <$ trace "XMonad recompiling because some files have changed."
+        else shouldCompile dirs CompileGhc
+shouldCompile _dirs CompileScript{} =
+    True <$ trace "XMonad recompiling because a custom build script is being used."
+
+getModTime :: FilePath -> IO (Maybe UTCTime)
+getModTime f = E.catch (Just <$> getModificationTime f) (\(SomeException _) -> return Nothing)
+
+-- | Compile the configuration.
+compile :: Directories -> Compile -> IO ExitCode
+compile dirs method =
+    bracket_ uninstallSignalHandlers installSignalHandlers $
+        bracket (openFile (errFileName dirs) WriteMode) hClose $ \err -> do
+            let run = runProc (cfgDir dirs) err
+            case method of
+                CompileGhc ->
+                    run "ghc" ghcArgs
+                CompileStackGhc stackYaml ->
+                    run "stack" ["build", "--silent", "--stack-yaml", stackYaml] .&&.
+                    run "stack" ("ghc" : "--stack-yaml" : stackYaml : "--" : ghcArgs)
+                CompileScript script ->
+                    run script [binFileName dirs]
+  where
+    ghcArgs = [ "--make"
+              , "xmonad.hs"
+              , "-i" -- only look in @lib@
+              , "-ilib"
+              , "-fforce-recomp"
+              , "-main-is", "main"
+              , "-v0"
+              , "-outputdir", buildDirName dirs
+              , "-o", binFileName dirs
+              ]
+
+    -- waitForProcess =<< System.Process.runProcess, but without closing the err handle
+    runProc cwd err exe args = do
+        hPutStrLn err $ unwords $ "$" : exe : args
+        hFlush err
+        (_, _, _, h) <- createProcess_ "runProc" (proc exe args){ cwd = Just cwd, std_err = UseHandle err }
+        waitForProcess h
+
+    cmd1 .&&. cmd2 = cmd1 >>= \case
+        ExitSuccess -> cmd2
+        e -> pure e
+
+-- | Check GHC output for deprecation warnings and notify the user if there
+-- were any. Report success otherwise.
+checkCompileWarnings :: Directories -> IO ()
+checkCompileWarnings dirs = do
+    ghcErr <- readFile (errFileName dirs)
+    if "-Wdeprecations" `isInfixOf` ghcErr
+      then do
+        let msg = unlines $
+                ["Deprecations detected while compiling xmonad config: " <> srcFileName dirs]
+                ++ lines ghcErr
+                ++ ["","Please correct them or silence using {-# OPTIONS_GHC -Wno-deprecations #-}."]
+        trace msg
+        xmessage msg
+      else
+        trace "XMonad recompilation process exited with success!"
+
+-- | Notify the user that compilation failed and what was wrong.
+compileFailed :: Directories -> ExitCode -> IO ()
+compileFailed dirs status = do
+    ghcErr <- readFile (errFileName dirs)
+    let msg = unlines $
+            ["Errors detected while compiling xmonad config: " <> srcFileName dirs]
+            ++ lines (if null ghcErr then show status else ghcErr)
+            ++ ["","Please check the file for errors."]
+    -- nb, the ordering of printing, then forking, is crucial due to
+    -- lazy evaluation
+    trace msg
+    xmessage msg
+
+-- | Recompile the xmonad configuration file when any of the following apply:
 --
---      * force is 'True'
+--  * force is 'True'
 --
---      * the xmonad executable does not exist
+--  * the xmonad executable does not exist
 --
---      * the xmonad executable is older than xmonad.hs or any file in
---        the @lib@ directory (under the configuration directory).
+--  * the xmonad executable is older than @xmonad.hs@ or any file in
+--    the @lib@ directory (under the configuration directory)
+--
+--  * custom @build@ script is being used
 --
 -- The -i flag is used to restrict recompilation to the xmonad.hs file only,
 -- and any files in the aforementioned @lib@ directory.
@@ -590,106 +733,21 @@ stateFileName = (</> "xmonad.state") <$> getXMonadDataDir
 --
 -- 'False' is returned if there are compilation errors.
 --
-recompile :: MonadIO m => Bool -> m Bool
-recompile force = io $ do
-    cfgdir  <- getXMonadDir
-    datadir <- getXMonadDataDir
-    let binn = "xmonad-"++arch++"-"++os
-        bin  = datadir </> binn
-        err  = datadir </> "xmonad.errors"
-        src  = cfgdir </> "xmonad.hs"
-        lib  = cfgdir </> "lib"
-        buildscript = cfgdir </> "build"
-
-    libTs <- mapM getModTime . Prelude.filter isSource =<< allFiles lib
-    srcT <- getModTime src
-    binT <- getModTime bin
-
-    useBuildscript <- do
-      exists <- doesFileExist buildscript
-      if exists
-        then do
-          isExe <- isExecutable buildscript
-          if isExe
-            then do
-              trace $ "XMonad will use build script at " ++ show buildscript ++ " to recompile."
-              return True
-            else do
-              trace $ unlines
-                [ "XMonad will not use build script, because " ++ show buildscript ++ " is not executable."
-                , "Suggested resolution to use it: chmod u+x " ++ show buildscript
-                ]
-              return False
-        else do
-          trace $
-            "XMonad will use ghc to recompile, because " ++ show buildscript ++ " does not exist."
-          return False
-
-    shouldRecompile <-
-      if useBuildscript || force
-        then return True
-        else if any (binT <) (srcT : libTs)
-          then do
-            trace "XMonad doing recompile because some files have changed."
-            return True
-          else do
-            trace "XMonad skipping recompile because it is not forced (e.g. via --recompile), and neither xmonad.hs nor any *.hs / *.lhs / *.hsc files in lib/ have been changed."
-            return False
-
-    if shouldRecompile
+recompile :: MonadIO m => Directories -> Bool -> m Bool
+recompile dirs force = io $ do
+    method <- detectCompile dirs
+    willCompile <- if force
+        then True <$ trace "XMonad recompiling (forced)."
+        else shouldCompile dirs method
+    if willCompile
       then do
-        -- temporarily disable SIGCHLD ignoring:
-        uninstallSignalHandlers
-        status <- bracket (openFile err WriteMode) hClose $ \errHandle ->
-            waitForProcess =<< if useBuildscript
-                               then compileScript bin cfgdir buildscript errHandle
-                               else compileGHC bin cfgdir errHandle
-
-        -- re-enable SIGCHLD:
-        installSignalHandlers
-
-        -- now, if it fails, run xmessage to let the user know:
+        status <- compile dirs method
         if status == ExitSuccess
-            then trace "XMonad recompilation process exited with success!"
-            else do
-                ghcErr <- readFile err
-                let msg = unlines $
-                        ["Error detected while loading xmonad configuration file: " ++ src]
-                        ++ lines (if null ghcErr then show status else ghcErr)
-                        ++ ["","Please check the file for errors."]
-                -- nb, the ordering of printing, then forking, is crucial due to
-                -- lazy evaluation
-                hPutStrLn stderr msg
-                forkProcess $ executeFile "xmessage" True ["-default", "okay", replaceUnicode msg] Nothing
-                return ()
-        return (status == ExitSuccess)
-      else return True
- where getModTime f = E.catch (Just <$> getModificationTime f) (\(SomeException _) -> return Nothing)
-       isSource = flip elem [".hs",".lhs",".hsc"] . takeExtension
-       isExecutable f = E.catch (executable <$> getPermissions f) (\(SomeException _) -> return False)
-       allFiles t = do
-            let prep = map (t</>) . Prelude.filter (`notElem` [".",".."])
-            cs <- prep <$> E.catch (getDirectoryContents t) (\(SomeException _) -> return [])
-            ds <- filterM doesDirectoryExist cs
-            concat . ((cs \\ ds):) <$> mapM allFiles ds
-       -- Replace some of the unicode symbols GHC uses in its output
-       replaceUnicode = map $ \c -> case c of
-           '\8226' -> '*'  -- •
-           '\8216' -> '`'  -- ‘
-           '\8217' -> '`'  -- ’
-           _ -> c
-       compileGHC bin dir errHandle =
-         runProcess "ghc" ["--make"
-                          , "xmonad.hs"
-                          , "-i"
-                          , "-ilib"
-                          , "-fforce-recomp"
-                          , "-main-is", "main"
-                          , "-v0"
-                          , "-o", bin
-                          ] (Just dir) Nothing Nothing Nothing (Just errHandle)
-       compileScript bin dir script errHandle =
-         runProcess script [bin] (Just dir) Nothing Nothing Nothing (Just errHandle)
+            then checkCompileWarnings dirs
+            else compileFailed dirs status
+        pure $ status == ExitSuccess
+      else
+        pure True
 
 -- | Conditionally run an action, using a @Maybe a@ to decide.
 whenJust :: Monad m => Maybe a -> (a -> m ()) -> m ()
