@@ -1,4 +1,4 @@
-{-# LANGUAGE MultiParamTypeClasses, FlexibleContexts #-}
+{-# LANGUAGE MultiParamTypeClasses, FlexibleContexts, NamedFieldPuns #-}
 ----------------------------------------------------------------------------
 -- |
 -- Module      :  XMonad.Main
@@ -16,7 +16,7 @@
 module XMonad.Main (xmonad, launch) where
 
 import System.Locale.SetLocale
-import qualified Control.Exception.Extensible as E
+import qualified Control.Exception as E
 import Data.Bits
 import Data.List ((\\))
 import Data.Function
@@ -24,7 +24,7 @@ import qualified Data.Map as M
 import qualified Data.Set as S
 import Control.Monad.Reader
 import Control.Monad.State
-import Data.Maybe (fromMaybe)
+import Data.Maybe (fromMaybe, isJust)
 import Data.Monoid (getAll)
 
 import Graphics.X11.Xlib hiding (refreshKeyboardMapping)
@@ -39,7 +39,7 @@ import XMonad.Operations
 import System.IO
 import System.Directory
 import System.Info
-import System.Environment
+import System.Environment (getArgs, getProgName, withArgs)
 import System.Posix.Process (executeFile)
 import System.Exit (exitFailure)
 import System.FilePath
@@ -48,6 +48,7 @@ import Paths_xmonad (version)
 import Data.Version (showVersion)
 
 import Graphics.X11.Xinerama (compiledWithXinerama)
+import Graphics.X11.Xrandr (xrrQueryExtension, xrrUpdateConfiguration)
 
 ------------------------------------------------------------------------
 
@@ -59,17 +60,17 @@ xmonad :: (LayoutClass l Window, Read (l Window)) => XConfig l -> IO ()
 xmonad conf = do
     installSignalHandlers -- important to ignore SIGCHLD to avoid zombies
 
+    dirs <- getDirectories
     let launch' args = do
-              catchIO buildLaunch
-              conf' @ XConfig { layoutHook = Layout l }
+              catchIO (buildLaunch dirs)
+              conf'@XConfig { layoutHook = Layout l }
                   <- handleExtraArgs conf args conf{ layoutHook = Layout (layoutHook conf) }
-              withArgs [] $ launch (conf' { layoutHook = l })
+              withArgs [] $ launch (conf' { layoutHook = l }) dirs
 
     args <- getArgs
     case args of
-        ("--resume": ws : xs : args') -> migrateState ws xs >> launch' args'
         ["--help"]            -> usage
-        ["--recompile"]       -> recompile True >>= flip unless exitFailure
+        ["--recompile"]       -> recompile dirs True >>= flip unless exitFailure
         ["--restart"]         -> sendRestart
         ["--version"]         -> putStrLn $ unwords shortVersion
         ["--verbose-version"] -> putStrLn . unwords $ shortVersion ++ longVersion
@@ -90,7 +91,7 @@ usage = do
         "Options:" :
         "  --help                       Print this message" :
         "  --version                    Print the version number" :
-        "  --recompile                  Recompile your ~/.xmonad/xmonad.hs" :
+        "  --recompile                  Recompile your xmonad.hs" :
         "  --replace                    Replace the running window manager with xmonad" :
         "  --restart                    Request a running xmonad process to restart" :
         []
@@ -111,21 +112,21 @@ usage = do
 --
 --   * Missing XMonad\/XMonadContrib modules due to ghc upgrade
 --
-buildLaunch ::  IO ()
-buildLaunch = do
+buildLaunch :: Directories -> IO ()
+buildLaunch dirs = do
     whoami <- getProgName
-    let compiledConfig = "xmonad-"++arch++"-"++os
+    let bin = binFileName dirs
+    let compiledConfig = takeFileName bin
     unless (whoami == compiledConfig) $ do
       trace $ concat
-        [ "XMonad is recompiling and replacing itself another XMonad process because the current process is called "
+        [ "XMonad is recompiling and replacing itself with another XMonad process because the current process is called "
         , show whoami
         , " but the compiled configuration should be called "
         , show compiledConfig
         ]
-      recompile False
-      dir  <- getXMonadDataDir
+      recompile dirs False
       args <- getArgs
-      executeFile (dir </> compiledConfig) False args Nothing
+      executeFile bin False args Nothing
 
 sendRestart :: IO ()
 sendRestart = do
@@ -134,7 +135,7 @@ sendRestart = do
     xmonad_restart <- internAtom dpy "XMONAD_RESTART" False
     allocaXEvent $ \e -> do
         setEventType e clientMessage
-        setClientMessageEvent e rw xmonad_restart 32 0 currentTime
+        setClientMessageEvent' e rw xmonad_restart 32 []
         sendEvent dpy rw False structureNotifyMask e
     sync dpy False
 
@@ -166,8 +167,8 @@ sendReplace = do
 -- function instead of 'xmonad'.  You probably also want to have a key
 -- binding to the 'XMonad.Operations.restart` function that restarts
 -- your custom binary with the resume flag set to @True@.
-launch :: (LayoutClass l Window, Read (l Window)) => XConfig l -> IO ()
-launch initxmc = do
+launch :: (LayoutClass l Window, Read (l Window)) => XConfig l -> Directories -> IO ()
+launch initxmc drs = do
     -- setup locale information from environment
     setLocale LC_ALL (Just "")
     -- ignore SIGPIPE and SIGCHLD
@@ -216,7 +217,9 @@ launch initxmc = do
             , buttonActions = mouseBindings xmc xmc
             , mouseFocused  = False
             , mousePosition = Nothing
-            , currentEvent  = Nothing }
+            , currentEvent  = Nothing
+            , directories   = drs
+            }
 
         st = XState
             { windowset       = initialWinset
@@ -231,7 +234,7 @@ launch initxmc = do
         runX cf st $ do
             -- check for serialized state in a file.
             serializedSt <- do
-                path <- stateFileName
+                path <- asks $ stateFileName . directories
                 exists <- io (doesFileExist path)
                 if exists then readStateFile initxmc else return Nothing
 
@@ -260,8 +263,11 @@ launch initxmc = do
 
             userCode $ startupHook initxmc
 
+            rrData <- io $ xrrQueryExtension dpy
+            let rrUpdate = when (isJust rrData) . void . xrrUpdateConfiguration
+
             -- main loop, for all you HOF/recursion fans out there.
-            forever $ prehandle =<< io (nextEvent dpy e >> getEvent e)
+            forever $ prehandle =<< io (nextEvent dpy e >> rrUpdate e >> getEvent e)
 
     return ()
       where
@@ -310,10 +316,15 @@ handle (MapRequestEvent    {ev_window = w}) = withDisplay $ \dpy -> do
 
 -- window destroyed, unmanage it
 -- window gone,      unmanage it
-handle (DestroyWindowEvent {ev_window = w}) = whenX (isClient w) $ do
+-- broadcast to layouts
+handle e@(DestroyWindowEvent {ev_window = w}) = do
+  whenX (isClient w) $ do
     unmanage w
     modify (\s -> s { mapped       = S.delete w (mapped s)
                     , waitingUnmap = M.delete w (waitingUnmap s)})
+  -- the window is already unmanged, but we broadcast the event to all layouts
+  -- to trigger garbage-collection in case they hold window-specific resources
+  broadcastMessage e
 
 -- We track expected unmap events in waitingUnmap.  We ignore this event unless
 -- it is synthetic or we are not expecting an unmap notification from a window.
@@ -472,8 +483,11 @@ grabKeys = do
     -- build a map from keysyms to lists of keysyms (doing what
     -- XGetKeyboardMapping would do if the X11 package bound it)
     syms <- forM allCodes $ \code -> io (keycodeToKeysym dpy code 0)
-    let keysymMap = M.fromListWith (++) (zip syms [[code] | code <- allCodes])
-        keysymToKeycodes sym = M.findWithDefault [] sym keysymMap
+    let keysymMap' = M.fromListWith (++) (zip syms [[code] | code <- allCodes])
+    -- keycodeToKeysym returns noSymbol for all unbound keycodes, and we don't
+    -- want to grab those whenever someone accidentally uses def :: KeySym
+    let keysymMap = M.delete noSymbol keysymMap'
+    let keysymToKeycodes sym = M.findWithDefault [] sym keysymMap
     forM_ (M.keys ks) $ \(mask,sym) ->
          forM_ (keysymToKeycodes sym) $ \kc ->
               mapM_ (grab kc . (mask .|.)) =<< extraModifiers
