@@ -1,4 +1,5 @@
 {-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE DeriveFunctor #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -52,6 +53,8 @@ module XMonad.StackSet (
     ) where
 
 import Prelude hiding (filter)
+import Control.Applicative.Backwards (Backwards (Backwards, forwards))
+import Data.Foldable (foldr, toList)
 import Data.Maybe   (listToMaybe,isJust,fromMaybe)
 import qualified Data.List as L (deleteBy,find,splitAt,filter,nub)
 import Data.List ( (\\) )
@@ -85,7 +88,7 @@ import qualified Data.Map  as M (Map,insert,delete,empty)
 --    continuation reified as a data structure.
 --
 --    The Zipper lets us replace an item deep in a complex data
---    structure, e.g., a tree or a term, without an  mutation.  The
+--    structure, e.g., a tree or a term, without a mutation.  The
 --    resulting data structure will share as much of its components with
 --    the old structure as possible.
 --
@@ -151,7 +154,7 @@ data Workspace i l a = Workspace  { tag :: !i, layout :: l, stack :: Maybe (Stac
     deriving (Show, Read, Eq)
 
 -- | A structure for window geometries
-data RationalRect = RationalRect Rational Rational Rational Rational
+data RationalRect = RationalRect !Rational !Rational !Rational !Rational
     deriving (Show, Read, Eq)
 
 -- |
@@ -175,8 +178,19 @@ data RationalRect = RationalRect Rational Rational Rational Rational
 data Stack a = Stack { focus  :: !a        -- focused thing in this set
                      , up     :: [a]       -- clowns to the left
                      , down   :: [a] }     -- jokers to the right
-    deriving (Show, Read, Eq)
+    deriving (Show, Read, Eq, Functor)
 
+instance Foldable Stack where
+    toList = integrate
+    foldr f z = foldr f z . toList
+
+instance Traversable Stack where
+    traverse f s =
+        flip Stack
+            -- 'Backwards' applies the Applicative in reverse order.
+            <$> forwards (traverse (Backwards . f) (up s))
+            <*> f (focus s)
+            <*> traverse f (down s)
 
 -- | this function indicates to catch that an error is expected
 abort :: String -> a
